import * as dgram from "dgram";
import { networkInterfaces } from "os";
import { convert_wave_file } from "./biw-utils.js";
export class AudioStreamTransmitter {
    num_channels;
    format;
    ptime;
    m_socket = null;
    prev_ts = null;
    sample_per_packet;
    seqnr;
    buf;
    wave;
    wave_pos = 0;
    wave_done;
    write_func;
    constructor(num_channels, format, ptime) {
        this.num_channels = num_channels;
        this.format = format;
        this.ptime = ptime;
        if (num_channels <= 0 || num_channels > 80 || num_channels != Math.round(num_channels))
            throw new Error("Invalid num_channels");
        this.sample_per_packet = (() => {
            switch (ptime) {
                case "p1": return 48;
                case "p0_666": return 32;
                case "p0_500": return 24;
                case "p0_333": return 16;
                case "p0_250": return 12;
                case "p0_125": return 6;
            }
        })();
        const bps = (() => {
            switch (format) {
                case "AM824": return 4;
                case "L24": return 3;
                case "L16": return 2;
            }
        })();
        const payload_size = bps * num_channels * this.sample_per_packet;
        if (payload_size > 1440)
            throw new Error("Payload size too big");
        this.buf = Buffer.alloc(12 + payload_size);
        this.buf.writeUInt8(0x80, 0);
        this.buf.writeUint32BE(Math.floor(Math.random() * (1 << 32)), 8);
        this.seqnr = Math.floor(Math.random() * 65536);
        this.write_func = (() => {
            const num_channels = this.num_channels;
            const buf = this.buf;
            switch (this.format) {
                case "AM824":
                    return (samples, off) => {
                        for (let ch = 0; ch < num_channels; ch++) {
                            const sample = (ch < samples.length) ? samples[ch] : 0;
                            buf.writeUInt32BE(0x30000000 | (sample & 0xFFFFFF), off);
                            off += 4;
                        }
                        return off;
                    };
                case "L24":
                    return (samples, off) => {
                        for (let ch = 0; ch < num_channels; ch++) {
                            const sample = (ch < samples.length) ? samples[ch] : 0;
                            buf.writeUIntBE(sample & 0xFFFFFF, off, 3);
                            off += 3;
                        }
                        return off;
                    };
                case "L16":
                    return (samples, off) => {
                        for (let ch = 0; ch < num_channels; ch++) {
                            const sample = (ch < samples.length) ? samples[ch] : 0;
                            buf.writeUInt16BE(sample & 0xFFFF, off);
                            off += 2;
                        }
                        return off;
                    };
            }
        })();
    }
    start(params) {
        if (this.m_socket)
            throw new Error("Already open");
        const src = params.src ?? (() => {
            const ifcs = networkInterfaces();
            for (const ifname of Object.keys(ifcs)) {
                for (const virtIfc of ifcs[ifname]) {
                    if (virtIfc.family == "IPv4" && !virtIfc.internal) {
                        return virtIfc.address;
                    }
                }
            }
            throw new Error("Could not find a local IPv4 interface");
        })();
        return new Promise((resolve, reject) => {
            const sentinel = setTimeout(() => reject(`Failed to open socket with local address ${src}:${params.port}`), 10000);
            const socket = dgram.createSocket({
                type: "udp4",
                reuseAddr: true
            });
            this.m_socket = socket;
            socket.on("error", (err) => reject(err));
            socket.bind({ address: src, port: params.port }, () => {
                const ttl = params.ttl ?? 64;
                socket.setMulticastInterface(src);
                socket.setMulticastTTL(ttl);
                socket.connect(params.port, params.dst, () => {
                    clearTimeout(sentinel);
                    const pt = params.pt ?? 96;
                    this.buf.writeUint8(pt, 1);
                    const sdp_ptime = (() => {
                        switch (this.ptime) {
                            case "p0_125": return "0.125";
                            case "p0_250": return "0.25";
                            case "p0_333": return "0.333";
                            case "p0_500": return "0.5";
                            case "p0_666": return "0.666";
                            case "p1": return "1.0";
                        }
                    })();
                    const name = params.name ?? "e2etest";
                    const sdp = `v=0
o=- 0 1 IN IP4 ${src}
s=${name}
i=${name}
t=0 0
m=audio ${params.port} RTP/AVP ${pt}
c=IN IP4 ${params.dst}/${ttl}
a=source-filter: incl IN IP4 ${params.dst} ${src}
a=rtpmap:${pt} ${this.format}/48000/${this.num_channels}
a=framecount:${this.sample_per_packet}
a=ptime:${sdp_ptime}
`;
                    resolve(sdp);
                });
            });
        });
    }
    stop() {
        const socket = this.m_socket;
        if (socket) {
            this.m_socket = null;
            socket.close();
        }
    }
    read_samples() {
        const wave = this.wave;
        if (wave == null)
            return [];
        const samples = wave.read_raw(this.wave_pos);
        this.wave_pos = this.wave_pos + 1;
        if (this.wave_pos >= wave.hdr.num_samples) {
            const done = this.wave_done;
            this.wave = undefined;
            this.wave_done = undefined;
            if (done)
                done();
        }
        return samples;
    }
    set_wave(wave, cb) {
        this.wave = convert_wave_file(wave, this.format == "L16" ? "L16" : "L24");
        this.wave_pos = 0;
        this.wave_done = cb;
    }
    process(_1, sync_seqnr, ts) {
        const socket = this.m_socket;
        if (!socket)
            return;
        if (this.prev_ts == null) {
            this.prev_ts = ts;
            return;
        }
        let cur_ts = this.prev_ts;
        let delta = (ts < cur_ts ? (0x100000000 + ts - cur_ts) : (ts - cur_ts)) >>> 0;
        if (delta >= 1000) {
            console.log(`Large jump is RTP-TS (${cur_ts} -> ${ts}) sync seq# ${sync_seqnr} tx seq# ${this.seqnr}`);
            this.prev_ts = ts;
            return;
        }
        let seqnr = this.seqnr;
        const sample_per_packet = this.sample_per_packet;
        const buf = this.buf;
        while (delta >= sample_per_packet) {
            buf.writeUint16BE(seqnr, 2);
            buf.writeUInt32BE(cur_ts, 4);
            let off = 12;
            for (let sidx = 0; sidx < sample_per_packet; sidx++) {
                const samples = this.read_samples();
                off = this.write_func(samples, off);
            }
            socket.send(buf);
            cur_ts = ((cur_ts + sample_per_packet) & 0xFFFFFFFF) >>> 0;
            seqnr = (seqnr + 1) & 0xFFFF;
            delta -= sample_per_packet;
        }
        this.seqnr = seqnr;
        this.prev_ts = cur_ts;
    }
    chain() {
        return (data, seqnr, ts) => this.process(data, seqnr, ts);
    }
}
